#
# [name] nkj.arg.py
# [exec] python -m nkj.arg <input.file> [-s <xsize>x<ysize>] [-p <xpitch>,<ypitch>]
#        python -m nkj.arg test
#        python -m nkj.arg test -p 0.01
#        python -m nkj.arg test -p 0.01,0.02 -s 1980x1024
#
# Written by Yoshikazu NAKAJIMA (Fri Mar 25 23:10:42 JST 2022)
#

import argparse
import re
from .str import *
from .math import *

argf_floatpair = lambda x:list(map(float, x.split(',')))

argf_size = lambda x:list(map(int, re.split('[xX]', x)))   # <文字列>.split() と re.split('<正規表現>', <文字列>) は違う関数．

def reshape_argumentpair(s):
	if (type(s) is not list):
		___ERROR___
	if (len(s) == 1): 
		s = [s[0], s[0]]
	elif (len(s) == 2): 
		pass
	else:
		___ERROR___
	return s

def reshape_argument(s):  # alias
	return reshape_argumentpair(s)


#-- main

if __name__ == '__main__':
	_DEBUG_LEVEL = 0
	_DEFAULT_IMAGESIZE = [640, 480]
	_DEFAULT_PIXELPITCH = [0.1, 0.1]

	debuglevel(_DEBUG_LEVEL)

	if (False):   #- preliminary test for lambda, list, map, and split().
		dprint((lambda x:list(map(int, x.split(','))))('100'))
		dprint((lambda x:list(map(int, x.split(','))))('100,200'))

	parser = argparse.ArgumentParser()

	parser.add_argument('input', help='input.file')
	parser.add_argument('-s', '--image_size', type=argf_size, default=_DEFAULT_IMAGESIZE, help='image size: <xsize>x<ysize>, ex) 640x480')
	parser.add_argument('-p', '--pixel_pitch', type=argf_floatpair, default=_DEFAULT_PIXELPITCH, help='pixel pitch')

	args = parser.parse_args()

	filename = args.input
	imagesize = args.image_size
	pixelpitch = reshape_argument(args.pixel_pitch)

	print('input file:   \'{0}\''.format(filename))
	print('image size:  {0}'.format(imagesize))
	print('pixel pitch: {0}'.format(pixelpitch))




#
# [name] nkj.cwt.py
# [exec] python -m nkj.cwt
#
# Written by Yoshikazu NAKAJIMA
#
_LIB_DEBUGLEVEL = 0

import os
import sys
import gc
from math import *
import numpy as np
from numpy import *
import matplotlib.pyplot as plt
import h5py as h5

sys.path.append(os.path.abspath(".."))
from nkj.str import *
from nkj.fa  import *
from nkj.filter import *
from nkj.func import *

_DEFAULT_WAVELET_SCALE = 1.0 # wavelet_scale = f / f_w
_DEFAULT_FREQUENCY = 1.0
_DEFAULT_ERROR_PADDING = 0

_WAVELET_DATA_BIAS_ELIMINATION_FLAG = False   # data bias elimination flag
_WAVELET_DATA_GAIN_NORMALIZATION_FLAG = False # data gain normalization flag

_WAVELET_SCALE = _DEFAULT_WAVELET_SCALE # wavelet scale
_WAVELET_GAUSS_SIGMA = 3.5 # Wavelet gauss sigma. It shows gauss function spread scale for sigma size.
_WAVELET_FUNCTION = "morlet" # {"morlet_wt", "morlet"}
_MIN_WAVELET_SCALE = 1.0
_MIN_WAVELET_GAUSS_SIGMA = 3.5

_ERROR_PADDING = _DEFAULT_ERROR_PADDING

def error_padding(padding=None):
	global _ERROR_PADDING
	if (padding == None):
		return _ERROR_PADDING
	else:
		_ERROR_PADDING = padding
		return True

def data_bias_elimination_flag(flag=None):
	global _WAVELET_DATA_BIAS_ELIMINATION_FLAG
	if (flag == None):
		return _WAVELET_DATA_BIAS_ELIMINATION_FLAG
	else:
		_WAVELET_DATA_BIAS_ELIMINATION_FLAG = flag
		return True

def data_gain_normalization_flag(flag=None):
	global _WAVELET_DATA_GAIN_NORMALIZATION_FLAG
	if (flag == None):
		return _WAVELET_DATA_GAIN_NORMALIZATION_FLAG
	else:
		_WAVELET_DATA_GAIN_NORMALIZATION_FLAG = flag
		return True

def wavelet_scale(s=None):
	global _WAVELET_SCALE
	if (s == None):
		return _WAVELET_SCALE
	else:
		if (s <= 0.0):
			print_error("Illegal wavelet scale.")
			return False
		elif (s < _MIN_WAVELET_SCALE):
			print_warning("Wavelet scale might be too small.")
		_WAVELET_SCALE = s
		return True

def wavelet_pseudo_frequency_scale(fw=None):
	if (fw == None):
		return 1.0 / wavelet_scale()
	else:
		wavelet_scale(1.0 / fw)
		return True

def wavelet_gauss_sigma(s=None): # DON'T CALL THIS FROM THE OUTSIDE IF NOT NEEDED.
	global _WAVELET_GAUSS_SIGMA
	if (s == None):
		return _WAVELET_GAUSS_SIGMA
	elif (s < _MIN_WAVELET_GAUSS_SIGMA):
		print_warning("Wavelet gauss sigma might be too small.")
	_WAVELET_GAUSS_SIGMA = s
	return True

def _wavelet_gauss_sigma_2(s=None): # DON'T CALL THIS FROM THE OUTSIDE IF NOT NEEDED.
	if (s == None):
		return 2.0 * wavelet_gauss_sigma()
	else:
		return wavelet_gauss_sigma(s / 2.0)

def _wavelet_gauss_scale(s=None): # This returns how many waves are included with the kernel.
	if (s == None):
		return wavelet_scale() * _wavelet_gauss_sigma_2()
	else:
		wavelet_scale(s / wavelet_gauss_sigma_2())
		return True

def wavelet_gauss_waves(s=None):
	return _wavelet_gauss_scale(s)

def wavelet_kernel_waves(s=None):
	return wavelet_gauss_waves(s)

def wavelet_frequency(f):
	return wavelet_scale() * f

def wave_frequency(fw): # fw: wavelet frequency
	return fw / wavelet_scale()

def wavelet_cycle(T): # T: cycle interval(time/length)
	return _wavelet_gauss_scale() * T

def wave_cycle(Tw): # Tw: wavelet cycle interval(time/length)
	return Tw / _wavelet_gauss_scale()

def wavelet_time(t): # t: time
	return t / _wavelet_gauss_scale()

def wavelet_gauss_time(t): # t: time
	return t / wavelet_scale()

def wave_time(tw): # tw: wavelet time
	return tw * wavelet_scale()

def morlet(x, f0): # x: time/length, f0: analyzed frequency
	gx = f0 * wavelet_gauss_time(x)
	g = exp(-sq(gx) / 2.0) # gauss
	wx = omega(f0) * x
	re = g * cos(wx)
	im = g * sin(wx)
	"""
	re = cos(wx)
	im = sin(wx)
	"""
	return re, im

def morlet_wt(wt): # wt: angle [-np.pi, np.pi]
	gx = wavelet_gauss_time(wt / (2.0 * np.pi))
	g = exp(-sq(gx) / 2.0) # gauss
	re = g * cos(wt)
	im = g * sin(wt)
	return re, im

def normalize_kernel(kernel_re, kernel_im):
	__VERSION = 2

	if (__VERSION == 1):
		# bias elimination

		dmean = mean([mean(kernel_re), mean(kernel_im)])
		kernel_re -= dmean
		kernel_im -= dmean

		dsum = 0.0
		for i in range(len(kernel_re)):
			dsum += sqrt(sq(kernel_re[i]) + sq(kernel_im[i]))
		dsum /= 2.0 # due to dividing to real and imaginary

		if (dsum != 0.0):
			kernel_re /= dsum
			kernel_im /= dsum

	elif (__VERSION == 2):
		#
		# kernel の re, im それぞれに対して、さらに +, - 成分をそれぞれで算出し、各々の重み和が 0.5 になるように補正する．
		#
		rpsum = 0.0  # real positive
		rnsum = 0.0  # real negative
		ipsum = 0.0  # imaginary positive
		insum = 0.0  # imaginary negative

		for i in range(len(kernel_re)):
			re = kernel_re[i]
			if (re >= 0.0):
				rpsum += re
			else:
				rnsum -= re   # += -re

		for j in range(len(kernel_im)):
			im = kernel_im[j]
			if (im >= 0.0):
				ipsum += im
			else:
				insum -= im   # += -im

		rpsum *= 2.0   # /= 0.5
		rnsum *= 2.0   # /= 0.5
		ipsum *= 2.0   # /= 0.5
		insum *= 2.0   # /= 0.5

		for i in range(len(kernel_re)):
			if (kernel_re[i] >= 0.0):
				if (rpsum != 0.0):
					kernel_re[i] /= rpsum
			else:
				if (rnsum != 0.0):
					kernel_re[i] /= rnsum

		for j in range(len(kernel_im)):
			if (kernel_im[j] >= 0.0):
				if (ipsum != 0.0):
					kernel_im[j] /= ipsum
			else:
				if (insum != 0.0):
					kernel_im[j] /= insum

	else: # other versions
		kernel_re = gain_normalization(bias_elimination(kernel_re))
		kernel_im = gain_normalization(bias_elimination(kernel_im))

	return kernel_re, kernel_im

def morlet_kernel(f, dt):
	#
	# f:  Analyzed frequency [Hz]
	# dt: Sampling interval(time/length)
	#
	ldprint2(["Analized frequency: ", f])

	fs = 1.0 / dt
	ldprint2(["Sampling frequency: ", fs])

	if (fs < 2.0 * f):
		print_error("Too low sampling frequency for the frequency.")
		ldprint2(["ERROR for frequency: ", float(f)])
		ldprint2(["ERROR: Fs: ", float(fs), " is lower than 2.0*F: ", 2.0 * float(f), " of F: ", float(f)])
		return None

	if (fs < 5.0 * f):
		print_warning("Low sampling frequency.")

	T = 1.0 / f
	ldprint2(["Cycle:         ", T])
	Tw = wavelet_cycle(T)
	ldprint2(["Wavelet scale: ", wavelet_scale()])
	ldprint2(["Wavelet cycle: ", Tw])
	ldprint2(["dt           : ", dt])
	kernelsize = refresh_size(Tw / dt + 1)
	ldprint2(["Kernel size: ", Tw / dt + 1, " -> ", kernelsize])

	footsize = foot_size(kernelsize)
	ldprint2(["Foot size:   ", footsize])

	kernel_re = []
	kernel_im = []
	for i in range(-footsize, footsize + 1):
		ldprint3(["Index: ", i])
		t = -(dt * i)
		ldprint3(["Time:  ", t])
		if (_WAVELET_FUNCTION == "morlet_wt"):
			wt = omega(f) * t
			re, im = morlet_wt(wt)
		elif (_WAVELET_FUNCTION == "morlet"):
			re, im = morlet(t, f)
		else:
			print_error("Illegal _WAVELET_FUNCTION specification.")
			os.sys.exit()
		kernel_re.append(re)
		kernel_im.append(im)

	kernel_re = np.array(kernel_re)
	kernel_im = np.array(kernel_im)

	# kernel normalization

	kernel_re, kernel_im = normalize_kernel(kernel_re, kernel_im)

	return kernel_re, kernel_im, kernelsize

def wavelet_f(data, dt, f): # dt: sampling interval [sec], f: analyzed frequency [Hz]
	ldprint2(["--> wavelet_f(data[", len(data), "], ", dt, ", ", f, ")"])

	datasize = len(data)
	ldprint2(["Data size: ", datasize])

	ret = morlet_kernel(f, dt)  # 周波数とサンプリング間隔（秒）からカーネルを生成

	if (ret == None):
		ldprint2(["<-- wavelet_f(): None"])
		zeroarray = np.zeros(datasize, dtype=np.float64)
		return False, zeroarray, zeroarray, zeroarray

	kr, ki, ks = ret
	ldprint2(["Filter size: ", ks])

	footsize = foot_size(ks)
	ldprint2(["Foot size:   ", footsize])

	offset = kernel_offset(ks)
	ldprint2(["Offset:      ", offset])

	wt = []
	wt_re = []
	wt_im = []

	index = 0
	while (index < offset):
		left = index - offset
		right = ks + left
		ldprint3(["Less[", index, "]: [", left, ", ", right - 1, "]"])
		ldata = data[-left : 0 : -1]
		rdata = data[0 : right]
		ldprint3(["Ldata size: ", len(ldata)])
		ldprint3(["Rdata size: ", len(rdata)])
		if (data_bias_elimination_flag()):
			signal = bias_elimination(np.array(concatenate([ldata, rdata])))
		else:
			signal = np.array(concatenate([ldata, rdata]))
		ldprint2(["Signal size: ", len(signal)])
		ldprint2(["Kernel size: ", ks])
		if (len(signal) != ks):
			'''
			print_error("Too short data compared with the analyzed frequency of {0:0.3f} Hz. Wavelet cycle length exceeded over the data size. Data are padded with {1:0.3f}.".format(f, error_padding()))
			'''
			dlen = len(data)
			return False, np.full(dlen, error_padding(), dtype=np.float64), np.full(dlen, error_padding(), dtype=np.float64), np.full(dlen, error_padding(), dtype=np.float64)
		re = np.dot(signal, kr)   # 信号とカーネルの内積をとる
		im = np.dot(signal, ki)   # 信号とカーネルの内積をとる
		wt.append(sqrt(sq(re) + sq(im)))   # 実部と虚部からパワーを計算
		wt_re.append(re)
		wt_im.append(im)
		index += 1
		if (data_bias_elimination_flag()):
			del signal
			gc.collect()

	while (index < datasize - offset):
		left = index - offset
		right = left + ks
		ldprint3(["Contained[", index, "]: [", left, ", ", right - 1, "]"])
		if (data_bias_elimination_flag()):
			signal = bias_elimination(data[left : right])
		else:
			signal = data[left : right]
		re = np.dot(signal, kr)
		im = np.dot(signal, ki)
		wt.append(sqrt(sq(re) + sq(im)))
		wt_re.append(re)
		wt_im.append(im)
		index += 1
		if (data_bias_elimination_flag()):
			del signal
			gc.collect()

	while (index < datasize):
		left = index - offset
		exceed = left + ks + - datasize
		right = datasize - 1 - exceed
		ldprint3(["Exceeding[", index, "]: [", left, ", +++", exceed, "(", right, ")]"])
		ldata = data[left : datasize]
		rdata = data[datasize - 1 : right : -1]
		ldprint3(["Ldata size:", len(ldata)])
		ldprint3(["Rdata size:", len(rdata)])
		if (data_bias_elimination_flag()):
			signal = bias_elimination(np.array(concatenate([ldata, rdata])))
		else:
			signal = np.array(concatenate([ldata, rdata]))
		re = np.dot(signal, kr)
		im = np.dot(signal, ki)
		wt.append(sqrt(sq(re) + sq(im)))
		wt_re.append(re)
		wt_im.append(im)
		index += 1
		if (data_bias_elimination_flag()):
			del signal
			gc.collect()

	ldprint2(["<-- wavelet_f()"])

	return True, np.array(wt), np.array(wt_re), np.array(wt_im)

def cwt_f(data, dt, f):
	return wavelet_f(data, dt, f)

def wavelet(data, dt, flist): # flist: frequency list
	ldprint(["--> wavelet(data[", len(data), "], ", dt, ")"])
	slist = []
	warray = np.empty((0, len(data)))
	rarray = np.empty((0, len(data)))
	iarray = np.empty((0, len(data)))
	for f in flist:
		ldprint(["f: ", f])
		status, wt, wt_re, wt_im = wavelet_f(data, dt, f)
		ldprint2(["Status: ", status])
		ldprint2(["Data[0]: ", wt.size])
		ldprint2(["Data[1]: ", wt_re.size])
		ldprint2(["Data[2]: ", wt_im.size])
		slist.append(status)
		warray = np.vstack((warray, wt))
		rarray = np.vstack((rarray, wt_re))
		iarray = np.vstack((iarray, wt_im))
	ldprint(["<-- wavelet()"])
	return slist, warray, rarray, iarray

def cwt(data, dt, flist):
	return wavelet(data, dt, flist)

def power_normalize_ws(data): # data: wavelet spectra
	ldprint(["--> power_normalize_ws()"])
	ldprint(["Data shape: ", data.shape])

	freqs = data.shape[0]
	samples = data.shape[1]
	ldprint(["Frequencies:  ", freqs])
	ldprint(["Data samples: ", samples])

	ndata = np.copy(data)
	for i in range(samples):
		sum = 0.0
		for fid in range(freqs):
			sum += sq(ndata[fid, i])
		if (sum != 0.0):
			sum = sqrt(sum)
			ldprint2(["Data norm: ", sum])
			ndata[:, i] /= sum

	ldprint(["<-- power_normalize_ws()"])
	return ndata

def statistics(data): # data: wavelet spectra
	ldprint(["--> statistics()"])
	ldprint(["Data shape: ", data.shape])
	freqs = data.shape[0]
	samples = data.shape[1]
	ldprint(["Frequencies:  ", freqs])
	ldprint(["Data samples: ", samples])
	spectra_av = np.mean(data, axis=1)  # axis=0: vertically(column-directionally) averaging, axis=1: horizontally(row-directionally) averaging
	spectra_sd = np.std(data, axis=1)
	ldprint(["<-- statistics()"])
	return spectra_av, spectra_sd, samples

def sd_normalize_array(data, biasadjustment=True):
	ldprint(["--> sd_normalize_array(, ", biasadjustment, ")"])
	if (biasadjustment is True):
		av = np.mean(data)
		data -= av
		sd = np.std(data)
	else:
		dlen = len(data)
		sd = 0.0
		if (dlen != 0):
			for i in range(len(data)):
				sd += (data[i])**2
			sd /= dlen
			sd = np.sqrt(sd)
	if (sd != 0.0):
		data /= sd
	ldprint(["<-- sd_normalize_array()"])
	return data

def sum_normalize_array(data, biasadjustment=True):
	if (biasadjustment is True):
		av = np.mean(data)
		data -= av
	sum = np.sum(np.abs(data))
	if (sum != 0.0):
		data /= sum
	data *= float(len(data))
	return data

def sd_abs_normalize_array(data):
	return sd_normalize_array(data, False)

def sum_abs_normalize_array(data): # DOESN'T ADJUST THE BIAS
	return sum_normalize_array(data, False)

def normalize_array(data):
	return sd_normalize_array(data)

def abs_normalize_array(data):
	return sd_abs_normalize_array(data)

def sequential_correlation(data, basesignal): # signal: signal data sequence, basesignal: a signal
	ldprint(["--> sequential_correlation()"])

	ldprint(["Data shape: ", data.shape])
	freqs = data.shape[0]
	samples = data.shape[1]
	ldprint(["Frequencies:  ", freqs])
	ldprint(["Data samples: ", samples])

	if (debuglevel() > 0):
		if (len(basesignal) != freqs):
			print_error("Illegal algorithm. ERROR#: NKJWAVELET-00402.")
			os.sys.exit()

	corr = []
	for i in range(samples):
		nsig = normalize_array(data[:, i])
		nbase = normalize_array(basesignal)
		corr.append(np.dot(nbase, nsig) / float(freqs))
		del nsig
		del nbase
		gc.collect()

	ldprint(["Corr data #:  ", len(corr)])
	ldprint(["<-- sequential_correlation()"])
	return np.array(corr)

def sequential_abs_correlation(data, basesignal): # signal: signal data sequence, basesignal: a signal
	ldprint(["--> sequential_abs_correlation()"])

	ldprint(["Data shape: ", data.shape])
	freqs = data.shape[0]
	samples = data.shape[1]
	ldprint(["Frequencies:  ", freqs])
	ldprint(["Data samples: ", samples])

	if (debuglevel() > 0):
		if (len(basesignal) != freqs):
			print_error("Illegal algorithm. ERROR#: NKJWAVELET-00426.")
			os.sys.exit()

	corr = []
	for i in range(samples):
		nsig = abs_normalize_array(data[:, i])
		nbase = abs_normalize_array(basesignal)
		corr.append(np.dot(nbase, nsig) / float(freqs))

	ldprint(["Corr data #:  ", len(corr)])
	ldprint(["<-- sequential_abs_correlation()"])
	return np.array(corr)

def sequential_average_correlation(data): # data: wavelet spectra
	av, sd, n = statistics(data)
	return sequential_correlation(data, av)

def sequential_average_abs_correlation(data): # data: wavelet spectra
	av, sd, n = statistics(data)
	return sequential_abs_correlation(data, av)

def temporal_correlation(data, base):
	return sequential_correlation(data, base)

def temporal_abs_correlation(data, base):
	return sequential_abs_correlation(data, base)

def temporal_average_correlation(data):
	return sequential_average_correlation(data)

def temporal_average_abs_correlation(data):
	return sequential_average_abs_correlation(data)

def correlate_ndarray(sig, base):
	ldprint(["--> correlate_ndarray()"])
	if (sig.shape != base.shape):
		ldprint0(["sig.shape:  ", sig.shape])
		ldprint0(["base.shape: ", base.shape])
		print_error("Data size mismatch. ERROR#: NKJWAVELET-00468.")
		return ValueError
	nsig = normalize_array(sig.flatten())
	nbase = normalize_array(base.flatten())
	corr = np.dot(nbase, nsig)
	datanum = len(nsig)
	if (datanum != 0):
		corr /= float(datanum)
	del nsig
	del nbase
	gc.collect()
	ldprint(["<-- correlate_ndarray()"])
	return corr

def correlate_phaseshifted_2darray(sig, base, phase):
	ldprint(["--> correlate_phaseshifted_2darray(,, ", phase, ")"])
	ldprint(["phase:  ", phase])
	if (sig.ndim != 2):
		print_error("Illegal data dimension. Data dimension is {0}. ERROR#: NKJWAVELET-00481.".format(sig.ndim))
		return ValueError
	if (base.ndim != 2):
		print_error("Illegal data dimension. Data dimension is {0}. ERROR#: NKJWAVELET-00486.".format(base.ndim))
		return ValueError
	ldprint(["sig.shape:  ", sig.shape])
	ldprint(["base.shape: ", base.shape])
	if (sig.shape != base.shape):
		print_error("Data size mismatch. ERROR#: NKJWAVELET-00495.")
		ldprint0(["sig.shape:  ", sig.shape])
		ldprint0(["base.shape: ", base.shape])
		return ValueError
	dlen = sig.shape[1]
	dt = phase
	ldprint(["dlen: ", dlen])
	ldprint(["dt:   ", dt])
	if (dt < -dlen or dlen - 1 < dt):
		print_error("Illegal phase. ERROR#: NKJWAVELET-00485.")
		return ValueError
	if (dt < 0):
		dlen = base.shape[1]
		csig = sig[:, -dt:]
		cbase = base[:, :dlen+dt]
		ldprint(["csig.shape:  ", csig.shape])
		ldprint(["cbase.shape: ", cbase.shape])
		if (debuglevel() > 0):
			if (csig.shape != cbase.shape):
				print_error("Illegal algorithm. ERROR#: NKJWAVELET-00492.")
				os.sys.exit()
	else:
		dlen = base.shape[1]
		csig = sig[:, :dlen-dt]
		cbase = base[:, dt:]
		ldprint(["csig.shape:  ", csig.shape])
		ldprint(["cbase.shape: ", cbase.shape])
		if (debuglevel() > 0):
			if (csig.shape != cbase.shape):
				print_error("Illegal algorithm. ERROR#: NKJWAVELET-00492.")
				os.sys.exit()
	corr = correlate_ndarray(csig, cbase)
	return corr

"""
def interval_correlation(data): # data: wavelet spectra
	ldprint(["--> interval_correlation()"])
	ldprint(["Data shape: ", data.shape])

	av, sd, n = statistics(data)

	freqs = data.shape[0]
	samples = data.shape[1]
	ldprint(["Frequencies:  ", freqs])
	ldprint(["Data samples: ", samples])

	ldprint(["<-- interval_correlation()"])
	return corr
"""

#-- main

if __name__ == '__main__':
	lib_debuglevel(1)

	# FLAGS

	FLAG_MORLET_TEST = False
	FLAG_WAVELET_KERNEL_TEST = True
	FLAG_WAVELET_TRANSFORM_TEST = True
	FLAG_WAVELET_TRANSFORM_TEST2 = True
	FLAG_2DARRAY_CORRELATION_TEST = False
	FLAG_PHASESHIFTED_2DARRAY_CORRELATION_TEST = False

	# CONSATNTS

	OPTIONS = "[analyzed frequency [wavelet_scale]]"
	SAMPLINGS = 300
	SAMPLING_INTERVAL =  0.1 # sampling interval [sec]
	ANALYZED_FREQUENCY = 2.0 # analyzed frequency [Hz]

	# args

	argc = len(sys.argv)
	ldprint(["Argc: ", argc])
	if (argc == 3):
		ANALYZED_FREQUENCY = float(sys.argv[1])
		if (wavelet_scale(float(sys.argv[2])) is False):
			os.sys.exit()
	elif (argc == 2):
		argv1 = sys.argv[1]
		if (argv1 == "help"):
			print_usage(OPTIONS)
			os.sys.exit()
		else:
			ANALYZED_FREQUENCY = float(sys.argv[1])
	elif (argc == 1):
		pass
	else:
		print_error("Illegal arguments.")

	# morlet test

	if (FLAG_MORLET_TEST):
		print("-- Morlet test.")

		xlist = []
		rlist = []
		ilist = []
		for x in np.arange(-0.5, 0.5, 0.001):
			xlist.append(x)
			if (_WAVELET_FUNCTION == "morlet_wt"):
				wt = omega(f) * t
				re, im = morlet_wt(omega(1.0) * x)
			elif (_WAVELET_FUNCTION == "morlet"):
				re, im = morlet(x, 1.0)
			else:
				print_error("Illegal _WAVELET_FUNCTION specification.")
				os.sys.exit()
			rlist.append(re)
			ilist.append(im)
		rlist = np.array(rlist)
		ilist = np.array(ilist)

		plt.plot(xlist, rlist, color='black')
		plt.plot(xlist, ilist, color='black', linestyle='dashed')
		plt.title("Morlet kernel")
		plt.xlabel("Time [sec]")
		plt.ylabel("Amplitude")
		plt.show()
		plt.close()

	# wavelet kernel test

	if (FLAG_WAVELET_KERNEL_TEST):
		print("-- wavelet kernel test.")

		ldprint(["Wavelet scale:       ", wavelet_scale()])
		ldprint(["Wavelet gauss sigma: ", wavelet_gauss_sigma()])
		ldprint(["Wavelet gauss waves: ", wavelet_gauss_waves()])

		dt = SAMPLING_INTERVAL

		ldprint(["Sampling frequency: ", 1.0 / dt])

		f = ANALYZED_FREQUENCY # analyzed frequenccy [Hz]
		ldprint(["Analyzed frequency: ", f])

		rkernel, ikernel, kernelsize = morlet_kernel(f, dt)
		ldprint(["Kernel size: ", kernelsize])

		footsize = foot_size(kernelsize)
		ldprint(["Foot size:   ", footsize])

		tlist = []
		for i in range(-footsize, footsize + 1):
			t = dt * i
			tlist.append(t)
		tlist = np.array(tlist)

		plt.plot(tlist, rkernel, color='black')
		plt.plot(tlist, ikernel, color='black', linestyle='dashed')
		plt.axvline(0.0)
		if (footsize * dt >= 1.0):
			plt.axvline(1.0)
		plt.title("Morlet kernel ({0:.3f} [Hz])".format(f))
		plt.xlabel("Time [sec]")
		plt.ylabel("Amplitude")
		plt.show()
		plt.close()

	# wavelet transform test

	if (FLAG_WAVELET_TRANSFORM_TEST):
		print("-- wavelet transform test.")

		dt = SAMPLING_INTERVAL
		data_samplings = SAMPLINGS
		f = ANALYZED_FREQUENCY # analyzed frequenccy [Hz]

		data = []
		for i in range(data_samplings):
			t = dt * i
			"""
			signal = sin(omega(2.0) * t) + cos(omega(5.0) * t)
			signal = cos(omega(f) * t)
			signal = 2.0 * sin(omega(0.2) * t) + cos(omega(f) * t) + cos(omega(3.0) * t) + 3.0 * cos(omega(5.0) * t)
			"""
			signal = 2.0 * sin(omega(0.2) * t) + cos(omega(3.0) * t) + 3.0 * cos(omega(5.0) * t)
			data.append(signal)
		data = np.array(data)

		status, wt, wt_re, wt_im = wavelet_f(data, dt, f)

		tlist = []
		for i in range(len(data)):
			t = dt * i
			tlist.append(t)
		tlist = np.array(tlist)

		# Plot

		fig = plt.figure()

		axis = []
		for i in range(2):
			axis.append(fig.add_subplot(2, 1, i + 1))

		axis[0].plot(tlist, data, color='black')
		axis[0].set_title("Orignal signal")
		axis[0].set_xlabel("Time [sec]")
		axis[0].set_ylabel("Amplitude")
		axis[1].plot(tlist, wt, color='black', linestyle='dotted')
		axis[1].plot(tlist, wt_re, color='black', linestyle='solid')
		axis[1].plot(tlist, wt_im, color='black', linestyle='dashed')
		axis[1].set_title("Wavelet spectra ({0:.3} [Hz])".format(f))
		axis[1].set_xlabel("Time [sec]")
		axis[1].set_ylabel("Spectra")
		plt.show()
		plt.close()

	# wavelet transform test #2

	if (FLAG_WAVELET_TRANSFORM_TEST2):
		print("-- wavelet transform test #2.")

		dt = SAMPLING_INTERVAL
		data_samplings = SAMPLINGS
		f = ANALYZED_FREQUENCY # analyzed frequenccy [Hz]

		data = []
		for i in range(data_samplings):
			t = dt * i
			"""
			signal = sin(omega(2.0) * t) + cos(omega(5.0) * t)
			signal = cos(omega(f) * t)
			"""
			signal = 2.0 * sin(omega(0.2) * t) + cos(omega(f) * t) + cos(omega(3.0) * t) + 3.0 * cos(omega(5.0) * t)
			data.append(signal)
		data = np.array(data)

		statuslist, wt, wt_r, wt_i = wavelet(data, dt, pseudo_logticks(0.01, 2.0)[::-1]) # RETURN: statuslist: status list, wt: wavelet power array, wt_r: wavelet real component array, wt_i: wavelet imaginary component array

	# 2D array correlation test

	if (FLAG_2DARRAY_CORRELATION_TEST):
		print("-- 2D array correlation test.")
		#with h5.File("./data/HDF5/Format/TMDU001_data_L.h5", 'r') as f:
		with h5.File("../data/resp/TMDU001_data_L.h5", 'r') as f:
			index = "{0}/{1}/{2}/Data".format(0, 'NOR', 0)
			print("Index:", index)
			d = f[index][()]
		corr = correlate_ndarray(d, d)
		print("Corr: ", corr)

	# Phase-shifted 2D array correlation test

	if (FLAG_PHASESHIFTED_2DARRAY_CORRELATION_TEST):
		print("-- phase-shifted 2D array correlation test.")
		#with h5.File("./waveletdata/TMDU001_L_NOR_0000_0000_ws.h5", 'r') as f:
		with h5.File("../data/resp/TMDU001_NOR_L_0000_0000_ws.h5", 'r') as f:
			d = f['Data'][()]
		for i in range(-50, 50):
			corr = correlate_phaseshifted_2darray(d, d, i)
			print("Corr: ", corr)

def cwtblur(f, x):  # f:基本周波数、x:blur成分を計算する周波数群
	_W = 0.1593923184378193
	_A = 0.7846546485469617
	_BLURSUM = 3.3997009862005765
	w = _W * f
	xc = f
	a = _A
	return extreme(x, w, xc, a)



#
# [name] nkj.os.py
#
# Written by Yoshikazu NAKAJIMA
#
_LIB_DEBUGLEVEL = 0

import os
import sys
import re
import glob

sys.path.append(os.path.abspath(".."))
from nkj.str import *

def parse_args(args):
	params = {}
	args = (' ' + ' '.join(args)).split(' -') 
	args = [a for a in args if a]   # remove blank elements

	for arg in args:
		split_arg = arg.split(' ')
		key = re.sub('^-*', '', split_arg[0])
		val = ' '.join(split_arg[1:])
		params[key] = val 

	return params

def get_filelist(dir=".", ext="*", refinement=None, elimination=None):
	ldprint(["Dir:    ", str_bracket(dir)])
	ldprint(["Ext:    ", str_bracket(ext)])
	filename = "*"
	if (type(refinement) is list):
		if (refinement != None):
			for refine in refinement:
				ldprint2(["Refine: ", str_bracket(refine)])
				filename += refine + "*"
	else:
		if (refinement != None):
			ldprint(["Refine: ", str_bracket(refinement)])
			filename += refinement + "*"
	if (ext[0] == '.'):
		filename += ext
	else:
		filename += "." + ext
	ldprint(["Filename: ", str_bracket(filename)])
	filelist = glob.glob(os.path.join(dir, filename))
	ldprint(["Filelist: ", filelist])
	if (elimination != None):
		newfilelist = []
		if (type(elimination) is list):
			for file in filelist:
				flag = True
				for elim in elimination:
					ldprint2(["Elim:   ", str_bracket(elim)])
					if (elim in file):
						flag = False
						break
				if (flag):
					newfilelist.append(file)
		else:
			for file in filelist:
				ldprint2(["Elim:   ", str_bracket(elimination)])
				if (elimination not in file):
					newfilelist.append(file)
		filelist = newfilelist
	filelist.sort()
	return filelist



#
# [name] nkj.file.py
#
# Written by Yoshiakzu NAKAJIMA
#
_LIB_DEBUGLEVEL = 0

import os
import sys
import json
import numpy as np

sys.path.append(os.path.abspath(".."))
from nkj.os import *
from nkj.np import *

def load_nparray(filename):
	try:
		return np.load(filename)
	except:
		return nparrayerrorval()

def save_nparray(filename, array):
	try:
		np.save(filename, array)
		return True
	except:
		return False

def save_nparraydict(filename, data):
	ldprint(["--> save_nparraydict(", str_bracket(filename), ", ", len(data), ")"])
	ldprint2(["Data type: ", type(data)])
	if (type(data) != dict):
		print_error("Illegal data type.")
		return False
	ndata = {}
	for key in data.keys():
		dprint2(["Key: ", str_bracket(key)])
		ndata[key] = str(tuple(data[key].tolist()))
	with open(filename, 'w') as f:
		json.dump(ndata, f, indent=2)
		return True
	return False

def _csvdictlist(origdict):
	dictlist = []
	for key in origdict:
		dprint2(["Dict key: \'{0}\'".format(key)])
		dict = origdict.get(key)
		if (dict != None):
			csvdict = {}
			csvdict.update(**{'Data label': key}, ** origdict.get(key))
			dictlist.append(csvdict)
	return dictlist

def _savecsv(filename, dictlist):
	try:
		with open(filename, 'w') as f:
			keylist = dictlist[0].keys()
			dprint(["Keys: {0}".format(keylist)])
			dprint(["Keys: {0}".format(list(keylist))])
			print("test")
			writer = csv.DictWriter(f, list(keylist), delimiter=',', quotechar='"')
			print("test")
			writer.writeheader()
			for rowdata in dictlist:
				dprint(["Row:  {0}".format(rowdata)])
				if (rowdata != None):
					writer.writerow(rowdata)
		return True
	except:
		return False

def savecsv(filename, dict):
	dictlist = _csvdictlist(dict)
	if (dictlist == None):
		return False
	if (True):
		dprint(["-- CSV dict list --"])
		for dict in dictlist:
			label = dict['Data label']
			dprint(["Dict[\'{0}\']: {1}".format(label, dict)])
		dprint(["--"])
	return _savecsv(filename, dictlist)

#-- main

if __name__ == '__main__':

	# Save test
	a = np.array([0, 1, 2, 3, 4, 5])
	print(a)

	testfile = "test.npy"

	if (not save_nparray(testfile, a)):
		print_error("Can't save \'{0}\'".format(testfile))
		os.sys.exit()

	# Load test
	b = load_nparray(testfile)
	if (is_nparrayerror(b)):
		print_error("Can't load \'{0}\'".format(testfile))
		os.sys.exit()
	print(b)



#
# [name] nkj.pygame.py
# [exec] python -m nkj.pygame
# [purpose] wrapping library for pygame
#
# Written by Yoshikazu NAKAJIMA
#

DEBUGLEVEL = 0

import os
import sys
import time
import cv2
import pygame

sys.path.append(os.path.abspath(".."))

import nkj.core as nk
import nkj.str as ns
from nkj.str import ldprint

import nkj.cv as ncv
from nkj.cv import *

class nkjgm():
	#-- class variables
	_classname = "nkjgm"

	def __init__(self):
		ldprint(["classname: ", nkjgm._classname])

#-- main

if __name__ == "__main__":
	WINDOWNAME = "image"
	LINEWIDTH = 5
	LINEBLEND = 2

	ns.LIB_DEBUGLEVEL = DEBUGLEVEL

	pygame.init() # PyGame の初期化

	cap = cv2.VideoCapture(0)

	# Initialize pygame image screen

	ret, frame = cap.read()

	height, width = frame.shape[:2]
	ldprint(["image size: (", str(width), ", ", str(height), ")"])

	screen = pygame.display.set_mode((width, height)) # ベーススクリーン

	# Initialize blended surface

	s = pygame.Surface((width, height)) # ブレンディッドサーフェイス
	s.set_alpha(64) # 0-255

	while True:
		ret, cvimage = cap.read()

		"""
		frame = cv2.hconcat([frame, frame]) # 横に連結
		frame = cv2.vconcat([frame, frame]) # 縦に連結
		"""
		"""
		# Image display with OpenCV
		cv2.imshow(WINDOWNAME, cvimage)
		if (cv2.waitKey(1) == 27):
			break
		"""

		# convert opencv image to pygame image
		rgbimage = cv2.cvtColor(cvimage, cv2.COLOR_BGR2RGB).swapaxes(0, 1)
		gmsurf = pygame.surfarray.make_surface(rgbimage)

		# display pygame image

		screen.blit(gmsurf, (0, 0)) # 指定（左上座標）の位置に画像を描画する

		# draw
		"""
		pygame.draw.aaline(screen, (255, 0, 0), (0, 0), (300, 300), LINEBLEND) # アンチエイリアシングを施した線の描画
		"""
		pygame.draw.line(screen, (255, 0, 0), (0, 0), (300, 300), LINEWIDTH) # 線の描画

		# alpha blending
		"""
		pygame.draw.polygon(s, (255, 255, 255), [(0, 0), (width, 0), (width, height), (0, height)]) # 全体を白で塗る
		pygame.draw.rect(s, (255, 255, 255), s.get_rect()) # 全体を白で塗る
		"""
		s.blit(gmsurf, (0, 0)) # 黒や白の背景ではカメラ画像が暗く（or 白っぽく）なるので、背景画像を描画．surface にも blit() は使える．
		pygame.draw.line(s, (255, 0, 0), (300, 0), (300, 300), LINEWIDTH) # 線の描画
		screen.blit(s, (0, 0)) # スクリーンにアルファブレンドする画像を描画する

		# flush
		pygame.display.update()
		"""
		pygame.display.flip() # ハードウェアアクセラレーションにも対応した update()
		"""

		"""
		pygame.event.get() # update() の後に、pygame プロセスイベントを発生させないとウィンドウが再描画されないことがあるので、プロセスを取得してみる
		"""

		flag = False
		for event in pygame.event.get():
			if (event.type == pygame.MOUSEBUTTONDOWN and event.button == 1):
				x, y = event.pos
				ldprint(["mouse click: (", str(x), ", ", str(y), ")"])

			if (event.type == pygame.QUIT):
				flag = True
				break
			if (event.type == pygame.KEYDOWN):
				if (event.key == pygame.K_ESCAPE):
					flag = True
					break
		if (flag):
			break

		"""
		key = pygame.key.get_pressed()
		if key[pygame.K_ESCAPE]:
			break
		"""

	pygame.quit() # PyGame の終了

	cap.release()
	cv2.destroyAllWindows()




#
# [name] nkj.np.py
#
# Written by Yoshiakzu NAKAJIMA
#
_LIB_DEBUGLEVEL = 0

import numpy as np

_nparrayerror = None

def nparrayerrorval(val=None):
	global _nparrayerror
	if (val is None):
		return _nparrayerror
	else:
		_nparrayerror = val
		return True

def is_nparrayerror(array):
	return (array.size == 0)

#-- main

if __name__ == '__main__':

	a = np.array([[1, 2], [2, 4], [3, 6]])
	print(a)

	covm = np.cov(a, rowvar=0, bias=1)
	print(covm)

	covm = np.cov(a, rowvar=0, bias=0)
	print(covm)



#
# [name] nkj.cv.py
# [exec] python cv.py
# [reference] https://qiita.com/teraken_/items/c8ac4c0ff60d8856a15c
#
# Written by Yoshikazu NAKAJIMA
#

DEBUGLEVEL = 2

import os
import sys
import cv2 as cv
import json
import sys
import platform

sys.path.append(os.path.abspath(".."))
import nkj.core as nk
import nkj.str as ns
from nkj.str import *

"""
OpenCV camera calibration で使用するデータの型について（参照：https://kamino.hatenablog.com/entry/opencv_calibrate_camera#sec4）

objectPoint:  vector<Point3f>         -> np.array([[x1, y1, z1], [x2, y2, z2], ...], dtype=np.float32)
objectPoints: vector<vector<Point3f>> -> [np.array([[x1, y1, z1], [x2, y2, z2], ...], dtype=np.float32)]  # モデルの姿勢ごとにnp.arrayにまとめてそれをリストにしている。
                                         [[np.array([x, y, z], dtype=np.float32)]] や、np.array([[[x1, y, z1], [x2, y2, z2], ...], ...], dtype=np.float32) では動作し"ない"ので注意。
imagePoint:   vector<Point2f>         -> np.array([[x1, y1], [x2, y2], ...], dtype=np.float32)
imagePoints:  vector<vector<Point2f>> -> [np.array([[x1, y1], [x2, y2], ...], dtype=np.float32)]  # モデルの姿勢ごとにnp.arrayにまとめてそれをリストにしている。
                                         [[np.array([x, y, z], dtype=np.float32)]] や、np.array([[[x1, y, z1], [x2, y2, z2], ...], ...], dtype=np.float32) では動作し"ない"ので注意。
"""

class nkjcv():
	#-- class variables
	_classname = "nkjcv"

	def __init__(self):
		#-- instance variables
		self._pt = []
		self._m = 0
		self._n = 0
		self._p = 0

	@classmethod
	def getClassName(cls):
		return nkjcv._classname

	def mouseevent_callback(self, event, x, y, flags, param):
		if (event == cv.EVENT_LBUTTONDOWN):
			self._m = self._n
			self._pt.append((x, y))
			self._n = self._n + 1
			self._m = self._m + 1
			ldpring(["pt[", str(self._n), "]: ", str(self._pt)])
		if (event == cv.EVENT_RBUTTONDOWN and self._n > 0):
			self._pt.pop(self._m - 1)
			self._m = self._m - 1
			self._n = self._n - 1
			self._p = self._p + 1
			ldpring(["pt[", str(self._n), "]: ", str(self._pt)])

	def clearPoints(self):
		self._pt.clear()
		self._m = 0
		self._n = 0

	def getPoints(self):
		return self._pt

	def getPoint(self, index):
		return self._pt[index]

	def getM(self):
		return self._m

	def getN(self):
		return self._n

	def getP(self):
		return self._p

#-- main

if __name__ == '__main__':
	ns.LIB_DEBUGLEVEL = DEBUGLEVEL

	WINDOW_TITLE = "image"

	nk.checkPythonVersion()

	ncv = nkjcv()

	cap = cv.VideoCapture(0) # カメラを設定
	cv.namedWindow(WINDOW_TITLE, cv.WINDOW_GUI_NORMAL) # ウィンドウの設定

	cv.setMouseCallback(WINDOW_TITLE, ncv.mouseevent_callback) # マウスイベントの設定

	print(">>> start!!")

	while True:
		ret, frame = cap.read()

		if (ncv.getN() >= 1):
			cv.drawMarker(frame, ncv.getPoint(ncv.getM() - 1), (0, 255, 255), markerType=cv.MARKER_TILTED_CROSS, markerSize=15)

		if (ncv.getN() >= 2):
			for i in range(1, ncv.getN()):
				cv.line(frame, ncv.getPoint(i - 1), ncv.getPoint(i), (0, 255, 255), 1)

		cv.imshow(WINDOW_TITLE, frame)

		k = cv.waitKey(1)

		if (k == 27):
			print(">>> Exit")
			break
		elif (k == ord("c")):
			print(">>> Clear coordinates.")
			ncv.clearPoints()
		elif (k == ord("s")):
			pirnt(">>> Save image and coordinates.")
			str = "patient(no:03}.png".format(no = ncv.getP())
			cv.imwrite(str, frame)
			with open("painted.json", "w") as f:
				json.dump(getPoints(), f, indent=4)

	cap.release()
	cv.destroyAllWindows()




#
# [name] nkj.prob.py
# [exec] python prob.py
#
# Written by Yoshikazu NAKAJIMA
#
_LIB_DEBUGLEVEL = 1

import os
import sys

sys.path.append(os.path.abspath(".."))
from nkj.str import *

def p_and(p_a, p_b):
	return p_a * p_b

def p_or(p_a, p_b):
	return (p_a + p_b - p_and(p_a, p_b))

def p_cap(p_a, p_b):
	return p_and(p_a, p_b)

def p_cup(p_a, p_b):
	return p_or(p_a, p_b)

def cap(p_a, p_b):
	return p_cap(p_a, p_b)

def cup(p_a, p_b):
	return p_cup(p_a, p_b)


#-- main

if __name__ == '__main__':
	lib_debuglevel(_LIB_DEBUGLEVEL)

	p_a = 0.3
	p_b = 0.7

	ldprint(["p_a: ", p_a])
	ldprint(["p_b: ", p_b])

	ldprint(["p_a AND p_b: ", p_and(p_a, p_b)])
	ldprint(["p_a OR p_b:  ", p_or(p_a, p_b)])



#
# [name] nkj.filter.py
# [exec] python -m nkj.filter
#
# Written by Yoshikazu NAKAJIMA
#

import os
import sys
from math import *
from numpy import *

sys.path.append(os.path.abspath(".."))
from nkj.math import *

def spread_size(foot_size):
	return int(ceil(foot_size)) * 2 + 1

def filter_size(foot_size):
	return spread_size(foot_size)

def foot_size(spread_size):
	return int(ceil((spread_size - 1.0) / 2.0))

def kernel_offset(spread_size):
	return foot_size(spread_size)

def refresh_size(size):
	return spread_size(foot_size(size))

def bias_elimination(data):
	dmean = mean(data)
	data -= dmean
	return data

def gain_normalization(data):
	dsum = sum(abs(data))
	if (dsum != 0.0):
		data /= dsum
	return data

def power_normalization(data):
	sd = sqrt(sum(np.power(data, 2)) / len(data)) # standard deviation
	if (sd != 0.0):
		data /= sd
	return data

class kernel():
	_classname = "nkj.filter.kernel"

	def __init__(self):
		#-- instance variables
		self._positivegain = 1.0
		self._negativegain = 1.0
		self._offset = 0.0

	@classmethod
	def getClassName(cls):
		return kernel._classname

	@classmethod
	@property
	def classname(cls):
		return cls.getClassName()

	def getPositiveGain(self):
		return self._positivegain

	def setPositiveGain(self, g):
		self._positivegain = g

	@property
	def positivegain(self):
		return self.getPositiveGain()

	@positivegain.setter
	def positivegain(self, g):
		self.setPositiveGain(g)

	@property
	def posgain(self):
		return self.getPositiveGain()

	@posgain.setter
	def posgain(self, g):
		self.setPositiveGain(g)

	@property
	def pg(self):
		return self.getPositiveGain()

	@pg.setter
	def pg(self, g):
		self.setPositiveGain(g)

	def getNegativeGain(self):
		return self._negativegain

	def setNegativeGain(self, g):
		self._negativegain = g

	@property
	def negativegain(self):
		return self.getNegativeGain()

	@negativegain.setter
	def negativegain(self, g):
		self.setNegativeGain(g)

	@property
	def neggain(self):
		return self.getNegativeGain()

	@neggain.setter
	def neggain(self, g):
		self.setNegativeGain(g)

	@property
	def ng(self):
		return self.getNegativeGain()

	@ng.setter
	def ng(self, g):
		self.setNegativeGain(g)

	def getGain(self):
		return sqrt((sq(self.getPositiveGain()) + sq(self.getNegativeGain())) / 2.0)

	def setGain(self, g):
		self._positivegain = g
		self._negativegain = g

	@property
	def gain(self):
		return self.getGain()

	@gain.setter
	def gain(self, gg):
		self.setGain(gg)

	@property
	def g(self):
		return self.getGain()

	@g.setter
	def g(self, gg):
		self.setGain(gg)

	def getOffset(self):
		return self._offset

	def setOffset(self, off):
		self._offset = off

	@property
	def offset(self):
		return self.getOffset()

	@offset.setter
	def offset(self, off):
		self.setOffset(off)

	@property
	def o(self):
		return self.offset

	@o.setter
	def o(self, off):
		self.setOffset(off)

	def getKernel(self):
		return None  # NOT implemented

	@property
	def kernel(self):
		return self.getKernel()

	@property
	def k(self):
		return self.getKernel()

#-- main

if __name__ == '__main__':

	if (True):
		filter_size = 7 
		foot = foot_size(filter_size)
		print("Foot size:", foot)
		print("Filter size:", filter_size, "->", spread_size(foot))

		filter_size = 8 
		foot = foot_size(filter_size)
		print("Foot size:", foot)
		print("Filter size:", filter_size, "->", spread_size(foot))

	if (True):
		k = kernel()
		print("Class name: \"{}\"".format(k.getClassName()))
		print("Class name: \"{}\"".format(k.classname))
		k.gain = 3.0
		print("Gain:   {}".format(k.getGain()))
		print("Gain:   {}".format(k.gain))
		print("Gain:   {}".format(k.g))
		k.offset = 2.5
		print("Offset: {}".format(k.getOffset()))
		print("Offset: {}".format(k.offset))
		print("Offset: {}".format(k.o))
		print("Kernel: {}".format(k.getKernel()))
		print("Kernel: {}".format(k.k))



#
# [name] nkj.serial.py
# [purpose] nkj.serial library
#
# Written by Yoshikazu NAKAJIMA (Wed Sep 23 14:38:26 JST 2020)
#

_DEBUGLEVEL = 0

CRLF = '\r\n'
CR = '\r'
LF = '\n'
EOB = '\0'

NONE = -1
DEFAULT_BAUDRATE = 9600
DEFAULT_TIMEOUT = 2.0
DEFAULT_SLEEPTIME = 0.1
DEFAULT_EOL = CRLF

import os
import sys
import serial
import time
from serial.tools import list_ports

sys.path.append(os.path.abspath(".."))
import nkj.str as ns
from nkj.str import *

if (ns.LIB_DEBUGLEVEL > 0):
	DEBUG = True
else:
	DEBUG = False

if (ns.LIB_DEBUGLEVEL > 1):
	DEBUG2 = True
else:
	DEBUG2 = False

class nkjserial:
	#-- Class variables
	_classname = "nkjserial"
	_comports = 0
	_comportlist = []
	_comport_prefix = None

	_baudrate = DEFAULT_BAUDRATE
	_timeout = DEFAULT_TIMEOUT
	_eol = DEFAULT_EOL

	_at_once = True
	#--

	def __init__(self, port=None, baudrate=-1, timeout=-1):
		ldprint(["--> nkjserial.__init__(", str_bracket(port), ")"])

		nkjserial._comports += 1

		#-- Instance variables
		self.portstr = None
		self._isopened = False

		self._timeout = nkjserial._timeout
		self._baudrate = nkjserial._baudrate

		self._eol = nkjserial._eol
		#--

		ldprint(["_at_cone: ", str(nkjserial._at_once)])

		if (nkjserial._at_once):
			nkjserial._at_once = False
			nkjserial._comportlist = nkjserial.getPortList()
			if (DEBUG):
				print("-- port list [", len(nkjserial._comportlist), "] --")
				for port in nkjserial._comportlist:
					print(port)
				print("--")

			nkjserial.getPortPrefix()

		if (ns.is_none(port)):
			return
		else:
			ldprint(["Port: ", str_bracket(port)])
			self.portstr = port

		status = self.open(port, baudrate, timeout)

		if (status):
			ldprint(["Serial opening is succeeded."])
		else:
			ldprint(["Serial opening is failed."])

		ldprint(["<-- nkjserial.__init__()"])

	def __del__(self):
		nkjserial._comports -= 1

		status = self.close()

		if (status):
			ldprint(["Serial closing is succeeded."])
		else:
			ldprint(["Serial closing is failed."])

	def __str__(self, title=None):
		mesg = None

		if (ns.is_none(title)):
			if (ns.is_none(self.portstr)):
				mesg = ns.concat(["--- ", "NONE", " ---", "\n"])
			else:
				mesg = ns.concat(["--- ", self.portstr, " ---", "\n"])
		else:
			mesg = ns.concat(["--- ", title, " ---", "\n"])
			if (ns.is_none(self.portstr)):
				mesg += "Com: NONE\n"
			else:
				mesg += ns.concat(["Com: ", self.portstr, "\n"])

		mesg += "Opened:   " + str(self.isOpened()) + "\n"
		mesg += "Baudrate: " + str(self.getBaudrate()) + "\n"
		mesg += "Timeout:  " + str(self.getTimeout()) + "\n"
		mesg += "Total ports: " + str(self.getPorts()) + "\n"
		mesg += "---"
		return mesg

	def _setPortNo(self, portno):
		self.portstr = nkjserial._comport_prefix + str(portno)

	def _sleep(self, sleeptime=DEFAULT_SLEEPTIME):
		time.sleep(sleeptime)

	def isOpened(self):
		return self._isopened

	@classmethod
	def getClassName(cls):
		return nkjserial._classname

	@classmethod
	def getClassBaudrate(cls):
		return nkjserial._baudrate

	def getBaudrate(self):
		return self._baudrate

	@classmethod
	def getClassTimeout(cls):
		return nkjserial._timeout

	def getTimeout(self):
		return self._timeout

	@classmethod
	def getClassEOL(cls):
		return nkjserial._eol

	def getEOL(self):
		return self._eol

	@classmethod
	def getClassTerminator(cls):
		return nkjserial.getClassEOL()

	def getTerminator(self):
		return self.getEOL()

	@classmethod
	def getPorts(cls):
		return nkjserial._comports

	def getPort(self):
		return self.portstr

	@classmethod
	def getPortList(cls, baudrate=-1, timeout=-1):
		ldprint(["--> nkjserial.getPortList()"])

		if (baudrate > 0):
			nkjserial._baudrate = baudrate

		if (timeout > 0):
			nkjserial._timeout = timeout

		if (len(nkjserial._comportlist) == 0):
			for comport in list_ports.comports():
				ldprint2(["Sys comport:        ", str(comport)])

				device = comport.device
				ldprint2(["Sys comport device: ", device])

				try:
					ser = serial.Serial(device, baudrate=nkjserial._baudrate, timeout=nkjserial._timeout)
					ldprint(["Opened comport device: ", str_bracket(device)])
					nkjserial._comportlist.append(device)
					ser.close()
				except:
					pass

		ldprint(["<-- nkjserial.getPortList()"])
		return nkjserial._comportlist

	@classmethod
	def getPortPrefix(cls):
		if (ns.is_none(nkjserial._comport_prefix)):
			comportlist = nkjserial.getPortList()

			if (len(comportlist) == 0):
				return None

			comport_str = comportlist[0]

			comport_prefix = None

			for i in range(len(comport_str)):
				charstr = comport_str[i]
				if (charstr.isdigit()):
					break
				else:
				 	comport_prefix = ns.concat([comport_prefix, charstr])

			ldprint(["comport_prefix: ", str_bracket(comport_prefix)])

			nkjserial._comport_prefix = comport_prefix

		return nkjserial._comport_prefix

	@classmethod
	def setClassBaudrate(cls, baudrate):
		nkjserial._baudrate = baudrate

	def setBaudrate(self, baudrate):
		self._baudrate = baudrate

	@classmethod
	def setClassTimeout(cls, timeout):
		nkjserial._timeout = timeout

	def setTimeout(self, timeout):
		self._timeout = timeout

	@classmethod
	def setClassEOL(cls, eol):
		nkjserial._eol = eol

	def setEOL(self, eol):
		self._eol = eol

	@classmethod
	def setClassTerminator(cls, terminator):
		nkjserial.setClassEOL(terminator)

	def setTerminator(self, terminator):
		self.setEOL(terminator)

	def setPort(self, port):
		self.portstr = port

	def open(self, port=None, baudrate=-1, timeout=-1):
		ldprint(["--> nkjserial.open(", str_bracket(port), ")"])

		if (self._isopened == True):
			ldprint(["ERROR: Com port has been opened."])
			return False

		if (ns.is_none(port)):
			port = self.portstr

		if (baudrate > 0):
			self._baudrate = baudrate

		if (timeout > 0):
			self._timeout = timeout

		try:
			self.ser = serial.Serial(port, baudrate=self._baudrate, timeout=self._timeout)
		except:
			return False

		self.portstr = port

		self._isopened = True

		self._sleep()

		ldprint(["<-- nkjserial.open()"])
		return True

	def connect(self, port=None, baudrate=-1, timeout=-1):
			return self.open(port, baudrate, timeout)

	def close(self):
		if (self._isopened == False):
			return False

		self.ser.close()
		return True

	def diconnect(self):
		return self.close()

	def clear(self):
		self.ser.reset_input_buffer()	# Flush input bffer, discarding all its contents. The old name is 'flushInput().'
		self.ser.reset_output_buffer()	# Clear output buffer, aborting the current output and discarding all that is in the buffer. The old name is 'flushOutput().'

	def flush(self):
		self.ser.flush()		# Flush of file like objects. In this case, wait until all data is written.

	def send(self, command):
		ldprint(["--> nkjserial.send(", str_bracket(command), ")"])

		if (self._isopened == False):
			ldprint(["ERROR: Serial port is not opened."])
			return False

		if (ns.is_none(command)):
			ldprint(["ERROR: Null command"])
			return False

		if (command == "\0"):
			ldprint(["ERROR: Null command"])
			return False

		if (len(command) < 1):
			ldprint(["ERROR: Command string is too short."])
			return False

		serstr = command + self.getEOL()

		ldprint(["Sent string:", str_bracket(serstr)])

		self.clear()

		self.ser.write(serstr.encode())

		self.flush()
		self._sleep()

		ldprint(["<-- nkjserial.send()"])
		return True

	def receive(self, chars=-1):
		ldprint(["--> nkjserial.receive(", str(chars), ")"])

		if (self._isopened == False):
			ldprint(["ERROR: Serial port is not opened."])
			return None

		if (chars < 0):
			#retstr = self.ser.read()
			retstr = self.ser.read_until(self.getEOL())
		else:
			retstr = self.ser.read(chars)

		self._sleep()

		msg = retstr.decode()

		ldprint(["Received message: ", str_bracket(msg)])

		msg = msg.strip(self.getEOL())

		ldprint(["Returning message: ", str_bracket(msg)])

		ldprint(["<-- nkjserial.receive()"])
		return msg

	def receiveline(self):
		ldprint(["--> nkjserial.receiveline()"])

		if (self._isopened == False):
			ldprint(["ERROR: Serial port is not opened."])
			return None

		retstr = self.ser.readline()

		self._sleep()

		msg = retstr.decode()

		ldprint(["Status: ", str_bracket(msg)])

		msg = msg.strip(self.getEOL())

		ldprint(["Returning message: ", str_bracket(msg)])

		ldprint(["<-- nkjserial.receiveline()"])
		return msg

#-- main

if __name__ == "__main__":
	ns.LIB_DEBUGLEVEL = _DEBUGLEVEL

	print("comport_prefix: ", str_bracket(nkjserial.getPortPrefix()))

	portlist = nkjserial.getPortList()

	print("-- port list [", len(portlist), "] --")
	for port in portlist:
		print(port)
	print("--")

	nkjserial.setClassEOL(CRLF)

	eol = nkjserial.getClassEOL()

	print("EOL: ", end=None)
	if (eol == '\r\n'):
		print("CRLF")
	elif (eol == '\r'):
		print("LF")
	elif (eol == '\n'):
		print("CR")
	else:
		print("unknown")

	ser = nkjserial("/dev/ttyS0")
	ser2 = nkjserial("/dev/ttyS1")

	print("Class name:", ser.getClassName())
	print("Ports:", ser.getPorts())
	print("Instance com port:", ser.getPort())
	print("Instance com port:", ser2.getPort())
	print("Is the instance opened:", ser.isOpened())
	print("Is the instance opened:", ser2.isOpened())

	print(ser)
	print(ser2)



#
# [name]    nkj.js.py
# [purpose] library for json
#
# Written by Yoshiakzu NAKAJIMA
#
_LIB_DEBUGLEVEL = 1

import os
import sys
import re
import numpy as np
import json

sys.path.append(os.path.abspath(".."))
import nkj.str as ns
from nkj.str import *

def load_nparrayjson(filename):
	with open(filename, 'r') as f:
		data = json.load(f)
		ndata = {}
		for key in data.keys():
			ndata[key] = ns.extract_float(data[key])
		return ndata
	return None

def save_nparrayjson(filename, data):
	if (type(data) != dict):
		print_error("Illegal data type.")
		return False
	ndata = {}
	for key in data.keys():
		ndata[key] = str(tuple(data[key].tolist()))
	with open(filename, 'w') as f:
		json.dump(ndata, f, indent=2)
		return True
	return False

#-- main

if __name__ == '__main__':
	USAGE = " input.json"

	lib_debuglevel(_LIB_DEBUGLEVEL)

	argc = len(sys.argv)

	if (argc == 2):
		filename = sys.argv[1]
	else:
		print_usage(sys.argv[0] + USAGE)
		os.sys.exit()
	ldprint(["Filename: ", str_bracket(filename)])

	data = load_nparrayjson(filename)
	if (data == None):
		print_error("Can't load input.json")
		os.sys.exit()

	ldprint(["Data: ", data])



#
# [name] setup.py
#
# Written by Yoshikazu NAKAJIMA
#

from setuptools import setup, find_packages

setup(
	name='nkj',
	version='1.01',
	description='Surgical Navigation Kernel and Job Control Library',
	author='Yoshikazu NAKAJIMA',
	license='bmi.tmdu',
	packages=find_packages()
)



#
# [name] nkj.str.py
# [purpose] nkj.str library
# [exec] python -m nkj.str
#
# Written by Yoshikazu NAKAJIMA (Wed Sep 23 14:38:26 JST 2020)
#

__DEBUGLEVEL = 0
__LIB_DEBUGLEVEL = 0

NULLSTR = ''
CR = '\r' # 行頭へ復帰
LF = '\n' # 改行
CRLF = '\r\n'

PRINTEND_CR = CR
PRINTEND_LF = LF
PRINTEND_CRLF = CRLF
PRINTEND_NONE = ''
DEFAULT_PRINTEND = PRINTEND_LF
DEFAULT_BRACKET = '\"'  # {'\'', '\"'}
DEFAULT_DOTPRINT = True

import os
import sys
import re
import numpy as np

_BRACKET = DEFAULT_BRACKET
_VERBOSE_DOTPRINT = DEFAULT_DOTPRINT
_VERBOSE_DOTPRINT_COUNTER = 0

def debuglevel(level=None):
	global __DEBUGLEVEL
	if (level is None):
		return __DEBUGLEVEL
	else:
		__DEBUGLEVEL = level
		return True

def lib_debuglevel(level=None):
	global __LIB_DEBUGLEVEL
	if (level is None):
		return __LIB_DEBUGLEVEL
	else:
		__LIB_DEBUGLEVEL = level
		return True

def verbose_dotprint(flag=None, interval=None):
	global _VERBOSE_DOTPRINT
	global _VERBOSE_DOTPRINT_COUNTER
	if (flag is None):
		if (_VERBOSE_DOTPRINT is True):
			if (interval == None):
				print(".", end='', file=sys.stderr)
				sys.stderr.flush()
			elif (_VERBOSE_DOTPRINT_COUNTER >= interval - 1):
				print(".", end='', file=sys.stderr)
				sys.stderr.flush()
				_VERBOSE_DOTPRINT_COUNTER = 0
			else:
				_VERBOSE_DOTPRINT_COUNTER += 1
	elif (flag == 'end'):
		print("", file=sys.stderr)
		sys.stderr.flush()
	elif (flag == 'flag'):
		return _VERBOSE_DOTPRINT
	elif (flag == 'enable'):
		_VERBOSE_DOTPRINT = True
	elif (flag == 'disable'):
		_VERBOSE_DOTPRINT = False
	else:
		_VERBOSE_DOTPRINT = flag
		return True

def bracket(type=None):
	global _BRACKET
	if (type is None):
		return _BRACKET
	else:
		if (type == 1):
			_BRACKET = '\''
		elif (type == 2):
			_BRACKET = '\"'
		else:
			print_error("Illegal bracket type. ERROR#: NKJSTR-00055.")

def str_bracket(in_str, type=1):
	if (in_str is None):
		s = bracket() + bracket()
	elif (in_str == ""):
		s = bracket() + bracket()
	else:
		s = bracket() + str(in_str) + bracket()
	return s

def str_float(in_str):
	return "{0:8.3f}".format(in_str)

def concat(strlist):
	retstr = NULLSTR
	if (len(strlist) != 0):
		for i in range(len(strlist)):
			if (strlist[i] != None and strlist[i] != NULLSTR):
				retstr += str(strlist[i])
	return retstr

def concat_path(pathlist):
	retstr = NULLSTR
	for path in pathlist[:-1]:
		dprint3(["PATH: ", str_bracket(path)])
		if (path[-1] != '/'):
			path += '/'
		retstr = concat([retstr, path])
	dprint3(["PATH: ", str_bracket(pathlist[-1])])
	retstr += pathlist[-1]
	return retstr

def print_usage(msg):
	print("Usage: python " + sys.argv[0] + " " + msg, flush=True)
	sys.stdout.flush()

def print_message(msg):
	print(msg, flush=True)
	sys.stdout.flush()

def print_warning(msg):
	print("WARNING: " + msg, flush=True)
	sys.stdout.flush()

def print_error(msg):
	print("ERROR: " + msg, flush=True)
	sys.stdout.flush()

def nparray_str(a:np.ndarray, title=None):
	mesg = NULLSTR
	if (title != None):
		mesg += "--- {} ---".format(title)
	"""
	# ■ ここに nparray の表示を実装すること
	s = shape(a)
	print("shape: {}".format(s))
	mesg += tuple(a)
	"""
	if (title != None):
		mesg += "---"
	return mesg

def print_nparray(a:np.ndarray, title=None):
	print(nparray_str(title))

def _dprint(level, strlist, end=DEFAULT_PRINTEND): # '_' で始まる関数は import * で読み込まれない
	if (__DEBUGLEVEL >= level):
		message = NULLSTR
		for i in range(len(strlist)):
			s = str(strlist[i])
			if (s != NULLSTR):
				message += s
			else:
				message += "NULL"
		print(message, flush=True, end=end)
		sys.stdout.flush()
	else:
		pass

def _ldprint(level, strlist, end=DEFAULT_PRINTEND):
	if (__LIB_DEBUGLEVEL >= level):
		message = NULLSTR
		for i in range(len(strlist)):
			s = str(strlist[i])
			if (s != NULLSTR):
				message += s
			else:
				message += "NULL"
		print(message, flush=True, end=end)
		sys.stdout.flush()
	else:
		pass

def dprint0(strlist, end=DEFAULT_PRINTEND):
	_dprint(0, strlist, end)

def dprint1(strlist, end=DEFAULT_PRINTEND):
	_dprint(1, strlist, end)

def dprint2(strlist, end=DEFAULT_PRINTEND):
	_dprint(2, strlist, end)

def dprint3(strlist, end=DEFAULT_PRINTEND):
	_dprint(3, strlist, end)

def dprint(strlist, end=DEFAULT_PRINTEND):
	dprint1(strlist, end)

def ldprint0(strlist, end=DEFAULT_PRINTEND):
	_ldprint(0, strlist, end)

def ldprint1(strlist, end=DEFAULT_PRINTEND):
	_ldprint(1, strlist, end)

def ldprint2(strlist, end=DEFAULT_PRINTEND):
	_ldprint(2, strlist, end)

def ldprint3(strlist, end=DEFAULT_PRINTEND):
	_ldprint(3, strlist, end)

def ldprint(strlist, end=DEFAULT_PRINTEND):
	ldprint1(strlist, end)

def is_none(str):
	if (str is None):
		return True
	elif (str == NULLSTR):
		return True
	else:
		return False

def isnot_none(str):
	return (not isNone(str))

def extract_float(istr):
	strs = re.findall(r"[-]*[0-9][0-9]*[.]*[0-9]*[e+-]*[0-9]*", istr)
	vals = []
	for s in strs:
		vals.append(float(s))
	return np.array(vals)

def atoi(s):
	return int(s.strip())

def linestr2int(s_line):
	dlist = []
	for s in s_line.split(','):
		dlist.append(atoi(s))
	return dlist

def l2i(linestr):
	return linestr2int(linestr)

def atof(s):
	return float(s.strip())

def linestr2float(s_line):
	linestr = s_line.split(',')
	if (len(linestr) == 1):
		linestr = s_line.split(' ')
	dlist = []
	for s in linestr:
		dlist.append(atof(s))
	return dlist

def l2f(linestr):
	return linestr2float(linestr)


#-- main

if __name__ == '__main__':
	import os
	import sys

	sys.path.append(os.path.abspath(".."))

	import nkj.core as nc
	print(["ROOTPATH: ", nc.rootpath()])

	import nkj.str as ns
	ns.debuglevel(5)
	ns.dprint(["test1, ", "test2, ", "test3"])

	"""
	debuglevel(5)
	"""
	print("test")
	__DEBUGLEVEL = 5
	dprint(["test, ", "test2, ", ""])

	print(concat_path(["test", "test2"]))
	print(concat_path(["test/", "test2"]))
	print(concat_path(["test/", "test2", "test3"]))

	print(linestr2float('   0.00123, 2.345,  -12.356'))
